import { Message } from "discord.js";
import { EsModuleClass, isModuleClass } from "src/util";

export interface CommandDocumentation {
    description: string;
    example: string;
}

export interface CommandOptions {
    pattern: string | RegExp;
    documentation?: CommandDocumentation;
    auth?: boolean;
}

export type BotAction = (actionsDone: boolean, m : Message, content: string) => boolean | Promise<boolean>;
export type BotMessageCommand = (message: Message, strippedContents: string, matches?: RegExpMatchArray) => void;

export interface ICommand {
    onStart?(): void | Promise<void>;
    _botCommands?: IBotCommand[];
    _botEvents?: { [action in ActionType]?: BotAction };
    BOT_COMMAND?: string;
}

export interface IBotCommand extends CommandOptions {
    action : BotMessageCommand;
}

export const BOT_COMMAND_DESCRIPTOR = "BOT_COMMAND";
export function CommandSet<T extends {new(...params: unknown[]): unknown}>(base: T): void {
    base.prototype.BOT_COMMAND = BOT_COMMAND_DESCRIPTOR;
}

export enum ActionType {
    MESSAGE,
    INDIRECT_MENTION,
    DIRECT_MENTION,
    POST_MESSAGE
}
export function Action(type: ActionType): MethodDecorator {
    return function<T>(target: unknown, propertyKey: string | symbol, descriptor: TypedPropertyDescriptor<T>): void {
        const command = target as ICommand;
        if(!command._botEvents)
            command._botEvents= {};

        command._botEvents[type] = descriptor.value as BotAction | undefined;
    };
}

export function Command(opts: CommandOptions) : MethodDecorator {
    return function<T>(target: unknown, propertyKey: string | symbol, descriptor: TypedPropertyDescriptor<T>): void {
        if(!descriptor.value)
            throw new Error("The decorator value must be initialized!");
        const command = target as ICommand;
        if(!command._botCommands)
            command._botCommands = [];

        command._botCommands.push({
            action: descriptor.value as unknown as BotMessageCommand,
            ...opts
        });
    };
}

export function isCommandSet(obj: unknown): obj is EsModuleClass<ICommand> {
    return isModuleClass<ICommand>(obj) && obj.prototype.BOT_COMMAND == BOT_COMMAND_DESCRIPTOR;
}